from typing import Any, Optional
from enum import Enum


class FMProperty():
    '''It defines a property of the feature model.
    
    Each property has a name, a description, and maybe a property parent.
    The property parent is used to organize the properties in a hierarchy.
    '''

    def __init__(self, 
                 name: str,  
                 description: Optional[str] = None,  
                 parent: Optional['FMProperty'] = None):
        self.name = name
        self.description = description
        self.parent = parent

    def to_dict(self) -> dict[str, Any]:
        return {'name': self.name, 
                'description': self.description,
                'parent': self.parent.name if self.parent is not None else None,
                'level': self.level()}

    def level(self) -> int:
        '''Return the level of parents.'''
        return 0 if self.parent is None else 1 + self.parent.level()


class FMPropertyMeasure():
    """It represent the measure of a property.
    
    It has the value of the measure, it can also have a size and a ratio.
    Example:
        property: Abstract Features.
        value: the list of abstract features.
        size: the length of the list.
        ratio: the percentage of abstract features with regards the total number of features.
    """
    def __init__(self, 
                 property: FMProperty,
                 value: Optional[Any] = None,  # Example: the list of abstract features
                 size: Optional[Any] = None,  # Example: number of abstract features
                 ratio: Optional[Any] = None):  # Example: percentage of abstract features from the total number of features
        self.property = property
        self.value = value
        self.size = size 
        self.ratio = ratio

    def to_dict(self) -> dict[str, Any]:
        result = {'value': safe_value(self.value),
                  'size': self.size,
                  'ratio': self.ratio}
        return self.property.to_dict() | result


class FMProperties(Enum):
    # METADATA
    NAME = FMProperty('Name', 'Name of the feature model.', None)
    DESCRIPTION = FMProperty('Description', 'Description of the feature model.', None)
    AUTHOR = FMProperty('Author', 'Author of the feature model', None)
    YEAR = FMProperty('Year', 'Year of creation of the feature model', None)
    REFERENCE = FMProperty('Reference', 'Main paper for reference or DOI of the feature model', None)
    TAGS = FMProperty('Tags', 'Tags or keywords that identify the feature model.', None)
    DOMAIN = FMProperty('Domain', 'Domain of the feature model.', None)
    LANGUAGE_LEVEL = FMProperty('Language level', 'Language level of the feature model.', None)

    # METRICS
    FEATURES = FMProperty('Features', 'Set of features in the feature model.', None)
    ABSTRACT_FEATURES = FMProperty('Abstract features', 'Features used to structure the feature model that, however, do not have any impact at implementation level.', FEATURES)
    CONCRETE_FEATURES = FMProperty('Concrete features', 'Features that are mapped to at least one implementation artifact.', FEATURES)
    LEAF_FEATURES = FMProperty('Leaf features', "Features that have not subfeatures (aka 'primitive features' or 'terminal features').", FEATURES)
    COMPOUND_FEATURES = FMProperty('Compound features', "Features that have subfeatures.", FEATURES)
    CONCRETE_COMPOUND_FEATURES = FMProperty('Concrete compound features', "Concrete and compound features.", CONCRETE_FEATURES)
    CONCRETE_LEAF_FEATURES = FMProperty('Concrete leaf features', "Concrete and leaf features.", CONCRETE_FEATURES)
    ABSTRACT_COMPOUND_FEATURES = FMProperty('Abstract compound features', "Abstract and compound features.", ABSTRACT_FEATURES)
    ABSTRACT_LEAF_FEATURES = FMProperty('Abstract leaf features', "Abstract and leaf features.", ABSTRACT_FEATURES)
    TREE_RELATIONSHIPS = FMProperty('Tree relationships', 'Number of relationships (edges) of the feature model.', None)
    ROOT_FEATURE = FMProperty('Root feature', 'The root of the feature model.', FEATURES)
    TOP_FEATURES = FMProperty('Top features', 'Features that are first descendants of the root.', ROOT_FEATURE)
    SOLITARY_FEATURES = FMProperty('Solitary features', 'Features that are not grouped in a feature group.', FEATURES)
    GROUPED_FEATURES = FMProperty('Grouped features', 'Features that occurs in a feature group.', FEATURES)
    MANDATORY_FEATURES = FMProperty('Mandatory features', 'Features marked as mandatory that need to be selected if its parent is selected.', TREE_RELATIONSHIPS)
    OPTIONAL_FEATURES = FMProperty('Optional features', 'Feature marked as optional.', TREE_RELATIONSHIPS)
    FEATURE_GROUPS = FMProperty('Feature groups', 'Features that express a choice over the grouped features in a group.', TREE_RELATIONSHIPS)
    ALTERNATIVE_GROUPS = FMProperty('Alternative groups', "Feature groups that require the selection of just one child (i.e., [1..1] cardinality).", FEATURE_GROUPS)
    OR_GROUPS = FMProperty('Or groups', "Feature groups that require the selection of at least one child (i.e., [1..*] cardinality).", FEATURE_GROUPS)
    MUTEX_GROUPS = FMProperty('Mutex groups', "Feature groups that require the selection of zero or just one child (i.e., [0..1] cardinality).", FEATURE_GROUPS)
    CARDINALITY_GROUPS = FMProperty('Cardinality groups', "Feature groups with arbitraty cardinality [a..b] that require the selection of an minimum and a maximum number of children.", FEATURE_GROUPS)
    BRANCHING_FACTOR = FMProperty('Branching factor', "Average number of children per non-leaf feature (aka 'Ratio of Variability').", None)  # Also 'Avg children per feature'
    MIN_CHILDREN_PER_FEATURE = FMProperty('Min children per feature', 'Minimal number of children per non-leaf feature.', BRANCHING_FACTOR)
    MAX_CHILDREN_PER_FEATURE = FMProperty('Max children per feature', 'Maximal number of children per feature.', BRANCHING_FACTOR)
    AVG_CHILDREN_PER_FEATURE = FMProperty('Avg children per feature', "Average number of children per feature.", BRANCHING_FACTOR)
    DEPTH_TREE = FMProperty('Depth of tree', 'Number of features of the longest path from the root to the leaf features.', None)
    MEAN_DEPTH_TREE = FMProperty('Mean depth of tree', 'Number of features of the mean path from the root to the leaf features.', DEPTH_TREE)
    FEATURE_ATTRIBUTES = FMProperty('Attributes', 'Features attributes in the model (i.e., number of distinct attributes).', None)
    FEATURES_WITH_ATTRIBUTES = FMProperty('Features with attributes', 'Features that contain some attributes defined in the model.', FEATURE_ATTRIBUTES)
    MIN_ATTRIBUTES_PER_FEATURE = FMProperty('Min attributes per feature', 'The minimal number of attributes in a feature.', FEATURE_ATTRIBUTES)
    MAX_ATTRIBUTES_PER_FEATURE = FMProperty('Max attributes per feature', 'The maximal number of attributes in a feature.', FEATURE_ATTRIBUTES)
    AVG_ATTRIBUTES_PER_FEATURE = FMProperty('Avg attributes per feature', 'Average number of attributes in features.', FEATURE_ATTRIBUTES)
    AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES = FMProperty('Avg attributes per feature w. attributes', 'Average number of attributes in features with attributes.', FEATURE_ATTRIBUTES)
    TYPED_FEATURES = FMProperty('Typed features', 'Non-Boolean features its selection require to provide a value (e.g., a number, a string,...', FEATURES)
    NUMERICAL_FEATURES = FMProperty('Numerical features', 'Features with a Integer or Real type.', TYPED_FEATURES)
    INTEGER_FEATURES = FMProperty('Integer features', 'Features with a Integer type.', NUMERICAL_FEATURES)
    REAL_FEATURES = FMProperty('Real features', 'Features with a Real type.', NUMERICAL_FEATURES)
    STRING_FEATURES = FMProperty('String features', 'Features with a String type.', TYPED_FEATURES)
    MULTI_FEATURES = FMProperty('Multi-features', "Features with cardinalities (aka 'clonable features')", FEATURES)

    CROSS_TREE_CONSTRAINTS = FMProperty('Cross-tree constraints', 'Textual cross-tree constraints.', None)
    LOGICAL_CONSTRAINTS = FMProperty('Logical constraints', 'Constraints with only logical operators.', CROSS_TREE_CONSTRAINTS)
    ARITHMETIC_CONSTRAINTS = FMProperty('Arithmetic constraints', 'Constraints with at least one arithmetic operator.', CROSS_TREE_CONSTRAINTS)
    AGGREGATION_CONSTRAINTS = FMProperty('Aggregation constraints', 'Constraints with at least one aggregation operator.', CROSS_TREE_CONSTRAINTS)
    SINGLE_FEATURE_CONSTRAINTS = FMProperty('Single feature constraints', 'Constraints with a single feature or negated feature.', LOGICAL_CONSTRAINTS)
    SIMPLE_CONSTRAINTS = FMProperty('Simple constraints', 'Requires and Excludes constraints.', LOGICAL_CONSTRAINTS)  # Requires and excludes
    REQUIRES_CONSTRAINTS = FMProperty('Requires constraints', 'Constraints modeling that the activation of a feature f1 implies the activation of a feature f2.', SIMPLE_CONSTRAINTS)
    EXCLUDES_CONSTRAINTS = FMProperty('Excludes constraints', 'Constraints modeling that two features are mutually exclusive and cannot be activated together.', SIMPLE_CONSTRAINTS)
    COMPLEX_CONSTRAINTS = FMProperty('Complex constraints', 'Constraints in arbitrary propositional logic formulae.', LOGICAL_CONSTRAINTS)  # Prop logic constraints (aka advanced constraints)
    PSEUDO_COMPLEX_CONSTRAINTS = FMProperty('Pseudo-complex constraints', 'Constraints that are convertible to a set of simple constraints.', COMPLEX_CONSTRAINTS)
    STRICT_COMPLEX_CONSTRAINTS = FMProperty('Strict-complex constraints', 'Constraints that cannot be converted to a set of simple constraints.', COMPLEX_CONSTRAINTS)
    MIN_CONSTRAINTS_PER_FEATURE = FMProperty('Min constraints per feature', 'The minimal number of constraints per feature.', CROSS_TREE_CONSTRAINTS)
    MAX_CONSTRAINTS_PER_FEATURE = FMProperty('Max constraints per feature', 'The maximal number of constraints per feature.', CROSS_TREE_CONSTRAINTS)
    AVG_CONSTRAINTS_PER_FEATURE = FMProperty('Avg constraints per feature', 'The average number of constraints per feature.', CROSS_TREE_CONSTRAINTS)
    EXTRA_CONSTRAINT_REPRESENTATIVENESS = FMProperty('Features in constraints', "Features involved in cross-tree constraints. The ratio to the total number of features is called 'Extra constraint representativeness (ECR)'.", CROSS_TREE_CONSTRAINTS)
    MIN_FEATURES_PER_CONSTRAINT = FMProperty('Min features per constraint', "The minimal number of features involved per cross-tree constraint.", EXTRA_CONSTRAINT_REPRESENTATIVENESS)
    MAX_FEATURES_PER_CONSTRAINT = FMProperty('Max features per constraint', "The maximal number of features involved per cross-tree constraint.", EXTRA_CONSTRAINT_REPRESENTATIVENESS)
    AVG_FEATURES_PER_CONSTRAINT = FMProperty('Avg features per constraint', "The average number of features involved per cross-tree constraint.", EXTRA_CONSTRAINT_REPRESENTATIVENESS)

    VALID = FMProperty('Satisfiable (valid)', 'A feature model is satisfiable (valid, not void) if it represents at least one configuration.', None)
    CORE_FEATURES = FMProperty('Core features', "Features that are part of all the configurations (aka 'common features').", None)  # Also 'Common features'
    DEAD_FEATURES = FMProperty('Dead features', 'Features that cannot appear in any configuration.', None)
    VARIANT_FEATURES = FMProperty('Variant features', 'Features that appear only in some configurations (i.e., features that are neither core nor dead).', None)  # Also 'Real optional features'
    UNIQUE_FEATURES = FMProperty('Unique features', 'Features that appear in exactly one configuration. The ratio is based on the total number of features.', VARIANT_FEATURES)
    PURE_OPTIONAL_FEATURES = FMProperty('Pure optional features', 'Feature with 0.5 (50%) probability of being selected in a valid configuration (i.e., their selection is unconstrained). The ratio is based on the total number of features.', VARIANT_FEATURES)
    FALSE_OPTIONAL_FEATURES = FMProperty('False-optional features', "Features included in all possible configurations although not being modelled as mandatory. The ratio is based on the total number of features.", CORE_FEATURES)
    CONFIGURATIONS = FMProperty('Configurations', 'Number of configurations represented by the feature model. If <= is shown, the number represents an upper estimation bound.', None)
    TOTAL_VARIABILITY = FMProperty('Total variability', 'The total variability measures the flexibility of the SPL considering all features.', None)
    PARTIAL_VARIABILITY = FMProperty('Partial variability', 'The partial variability measures the flexibility of the SPL considering only variant features.', None)
    HOMOGENEITY = FMProperty('Homogeneity', 'The homogeneity measures how similar are the configurations of the SPL.', None)
    PRODUCT_DISTRIBUTION = FMProperty('Configuration distribution', 'Number of configurations having a given number of features.', None)
    PD_MODE = FMProperty('Mode', 'Most frequently occurring number of features for a configuration.', PRODUCT_DISTRIBUTION)
    PD_MIN = FMProperty('Min', 'Number of features in the smallest configuration.', PRODUCT_DISTRIBUTION)
    PD_MAX = FMProperty('Max', 'Number of features in the largest configuration.', PRODUCT_DISTRIBUTION)
    PD_RANGE = FMProperty('Range', 'Range of number of features in configurations.', PRODUCT_DISTRIBUTION)
    PD_MEAN = FMProperty('Mean', 'Mean number of features occurring in configurations.', PRODUCT_DISTRIBUTION)
    PD_STD = FMProperty('Standard deviation', 'Standard deviation of number of features in configurations.', PRODUCT_DISTRIBUTION)
    PD_MEDIAN = FMProperty('Median', 'Median number of features in configurations.', PRODUCT_DISTRIBUTION)
    PD_MAD = FMProperty('Median absolute deviation', 'Median absolute deviation number of features in configurations.', PRODUCT_DISTRIBUTION)
    
    # ATOMIC_SETS = FMProperty('Atomic sets', '', None)  # Atomic sets need to be fixed in FLAMA.


def safe_value(value: Any) -> str:
    if isinstance(value, str):
        return value.replace('"', '')
    if isinstance(value, list):
        return [safe_value(v) for v in value if isinstance(v, str)]
    return value



"""
This module contains all utils related to the management of cross-tree constraints.
"""

from flamapy.core.models import AST, ASTOperation
from flamapy.core.models import ast as ast_utils
from flamapy.metamodels.fm_metamodel.models import Constraint


def is_simple_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is a simple constraint (requires or excludes)."""
    return is_requires_constraint(constraint) or is_excludes_constraint(constraint)


def is_complex_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is a complex constraint 
    (i.e., it is not a simple constraint)."""
    return not is_simple_constraint(constraint)


def is_requires_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is a requires constraint."""
    root_op = constraint.ast.root
    if root_op.is_binary_op():
        if root_op.data in [ASTOperation.REQUIRES, ASTOperation.IMPLIES]:
            return root_op.left.is_term() and root_op.right.is_term()
        elif root_op.data == ASTOperation.OR:
            neg_left = root_op.left.data == ASTOperation.NOT and root_op.left.left.is_term()
            neg_right = root_op.right.data == ASTOperation.NOT and root_op.right.left.is_term()
            return neg_left and root_op.right.is_term() or neg_right and root_op.left.is_term()
    return False


def is_excludes_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is an excludes constraint."""
    root_op = constraint.ast.root
    if root_op.is_binary_op():
        if root_op.data in [ASTOperation.EXCLUDES, ASTOperation.XOR]:
            return root_op.left.is_term() and root_op.right.is_term()
        elif root_op.data in [ASTOperation.REQUIRES, ASTOperation.IMPLIES]:
            neg_right = root_op.right.data == ASTOperation.NOT and root_op.right.left.is_term()
            return root_op.left.is_term() and neg_right
        elif root_op.data == ASTOperation.OR:
            neg_left = root_op.left.data == ASTOperation.NOT and root_op.left.left.is_term()
            neg_right = root_op.right.data == ASTOperation.NOT and root_op.right.left.is_term()
            return neg_left and neg_right
    return False


def is_pseudo_complex_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is a pseudo-complex constraint 
    (i.e., it can be transformed to a set of simple constraints)."""
    split_ctcs = split_constraint(constraint)
    return len(split_ctcs) > 1 and all(is_simple_constraint(ctc) for ctc in split_ctcs)


def is_strict_complex_constraint(constraint: Constraint) -> bool:
    """Return true if the constraint is a strict-complex constraint 
    (i.e., it cannot be transformed to a set of simple constraints)."""
    split_ctcs = split_constraint(constraint)
    return any(is_complex_constraint(ctc) for ctc in split_ctcs)


def get_new_ctc_name(ctcs_names: list[str], prefix_name: str) -> str:
    """Return a new name for a constraint (based on the provided prefix) that is not already 
    in the given list of constraints' names."""
    count = 1
    new_name = f'{prefix_name}'
    while new_name in ctcs_names:
        new_name = f'{prefix_name}{count}'
        count += 1
    return new_name


def left_right_features_from_simple_constraint(simple_ctc: Constraint) -> tuple[str, str]:
    """Return the names of the features involved in a simple constraint.
    
    A simple constraint can be a requires constraint or an excludes constraint.
    A requires constraint can be represented in the AST of the constraint with one of the 
    following structures:
        A requires B
        A => B
        !A v B
    An excludes constraint can be represented in the AST of the constraint with one of the 
    following structures:
        A excludes B
        A => !B
        !A v !B
    """
    root_op = simple_ctc.ast.root
    if root_op.data in [ASTOperation.REQUIRES, ASTOperation.IMPLIES, ASTOperation.EXCLUDES]:
        left = root_op.left.data
        right = root_op.right.data
        if right == ASTOperation.NOT:
            right = root_op.right.left.data
    elif root_op.data == ASTOperation.OR:
        left = root_op.left.data
        right = root_op.right.data
        if left == ASTOperation.NOT and right == ASTOperation.NOT:  # excludes
            left = root_op.left.left.data
            right = root_op.right.left.data
        elif left == ASTOperation.NOT:  # implies: A -> B
            left = root_op.left.left.data
            right = root_op.right.data
        elif right == ASTOperation.NOT:  # implies: B -> A
            left = root_op.right.left.data
            right = root_op.left.data
    return (left, right)


def split_constraint(constraint: Constraint) -> list[Constraint]:
    """Given a constraint, split it in multiple constraints separated by the AND operator."""
    asts = split_formula(constraint.ast)
    asts_simplified = [ast_utils.simplify_formula(ast) for ast in asts]
    asts = []
    for ctc in asts_simplified:
        asts.extend(split_formula(ctc))
        
    asts_negation_propagated = [ast_utils.propagate_negation(ast.root) for ast in asts]
    asts = []
    for ctc in asts_negation_propagated:
        asts.extend(split_formula(ctc))

    asts_cnf = [ast_utils.to_cnf(ast) for ast in asts]
    asts = []
    for ctc in asts_cnf:
        asts.extend(split_formula(ctc))
    return [Constraint(f'{constraint.name}{i}', ast) for i, ast in enumerate(asts)]


def split_formula(formula: AST) -> list[AST]:
    """Given a formula, returns a list of formulas separated by the AND operator."""
    res = []
    node = formula.root
    if node.data == ASTOperation.AND:
        res.extend(split_formula(AST(node.left)))
        res.extend(split_formula(AST(node.right)))
    else:
        res.append(formula)
    return res


import json
import tempfile
from urllib.parse import urlparse
import pathlib
from typing import Any
import urllib.request

from flamapy.core.exceptions import FlamaException
from flamapy.metamodels.fm_metamodel.models import FeatureModel
from flamapy.metamodels.fm_metamodel.transformations import (
    UVLReader,
    FeatureIDEReader,
    AFMReader,
    GlencoeReader,
    JSONReader
)

from fmfactlabel import FMProperty, FMAnalysis, FMMetadata, FMMetrics


SPACE = ' '
INDENT_MULTIPLIER = 1  # change to 2 if you need more indentation


class FMCharacterization():
    
    def __init__(self, model: FeatureModel, light_fact_label: bool = False) -> None:
        self.metadata = FMMetadata(model)
        self.metrics = FMMetrics(model)
        self.analysis = FMAnalysis(model, light_fact_label)
    
    @staticmethod
    def from_path(fm_filepath: str, light_fact_label: bool = False) -> 'FMCharacterization':
        """Load characterization from a feature model file."""
        fm_model = read_fm_file(fm_filepath)
        characterization = FMCharacterization(fm_model, light_fact_label)
        characterization.metadata.name = fm_filepath.split('.')[0]
        return characterization

    @staticmethod
    def from_url(fm_url_filepath: str, light_fact_label: bool = False) -> 'FMCharacterization':
        """Load characterization from a feature model URL."""
        with tempfile.NamedTemporaryFile(suffix=".uvl", mode='w+', delete=True) as tmp:
            urllib.request.urlretrieve(fm_url_filepath, tmp.name)
            characterization = FMCharacterization.from_path(tmp.name, light_fact_label)
            characterization.metadata.name = get_filename_from_url(fm_url_filepath)
            return characterization
    
    def clean(self) -> None:
        self.analysis.clean()

    def __str__(self) -> str:
        lines = ['METADATA']
        for property in self.metadata.get_metadata():
            name = property.property.name
            value = str(property.value)
            lines.append(f'{name}: {value}')    

        lines.append('METRICS')
        for property in self.metrics.get_metrics():
            indentation = SPACE * get_parents_numbers(property.property)
            name = property.property.name
            value = str(property.value) if property.size is None else str(property.size)
            ratio = f' ({str(property.ratio*100)}%)' if property.ratio is not None else ''
            lines.append(f'{indentation}{name}: {value}{ratio}')    
        
        lines.append('ANALYSIS')
        for property in self.analysis.get_analysis():
            indentation = SPACE * get_parents_numbers(property.property)
            name = property.property.name
            value = str(property.value) if property.size is None else str(property.size)
            ratio = f' ({str(property.ratio*100)}%)' if property.ratio is not None else ''
            lines.append(f'{indentation}{name}: {value}{ratio}')    
        return '\n'.join(lines)

    @staticmethod
    def json_to_text(data: dict) -> str:
        lines = ['METADATA']
        for prop in data.get("metadata", []):
            name = prop.get("name")
            value = str(prop.get("value", "None"))
            lines.append(f"{name}: {value}")

        lines.append('METRICS')
        for prop in data.get("metrics", []):
            indentation = SPACE * (prop.get("level", 0) * INDENT_MULTIPLIER)
            name = prop.get("name")
            value = str(prop["size"]) if prop.get("size") is not None else str(prop.get("value"))
            ratio = prop.get("ratio")
            ratio_str = f" ({round(ratio * 100, 2)}%)" if ratio is not None else ""
            lines.append(f"{indentation}{name}: {value}{ratio_str}")

        lines.append('ANALYSIS')
        for prop in data.get("analysis", []):
            indentation = SPACE * (prop.get("level", 0) * INDENT_MULTIPLIER)
            name = prop.get("name")
            value = str(prop["size"]) if prop.get("size") is not None else str(prop.get("value"))
            ratio = prop.get("ratio")
            ratio_str = f" ({round(ratio * 100, 2)}%)" if ratio is not None else ""
            lines.append(f"{indentation}{name}: {value}{ratio_str}")

        return "\n".join(lines)

    def to_json(self) -> dict[Any]:
        metadata = []
        metrics = []
        analysis = []

        for property in self.metadata.get_metadata():
            metadata.append(property.to_dict())

        for property in self.metrics.get_metrics():
            metrics.append(property.to_dict())

        for property in self.analysis.get_analysis():
            analysis.append(property.to_dict())

        result = {}
        result['metadata'] = metadata
        result['metrics'] = metrics
        result['analysis'] = analysis
        return result

    def to_json_str(self) -> str:
        result = self.to_json()
        return json.dumps(result, indent=4)

    def to_json_file(self, filepath: str = None) -> None:
        result = self.to_json()
        with open(filepath, 'w', encoding='utf-8') as output_file:
            json.dump(result, output_file, indent=4)


def get_parents_numbers(property: FMProperty) -> int:
    if property.parent is None:
        return 1
    return 1 + get_parents_numbers(property.parent)


def read_fm_file(filename: str) -> FeatureModel | None:
    try:
        if filename.endswith(".uvl"):
            return UVLReader(filename).transform()
        elif filename.endswith(".xml") or filename.endswith(".fide"):
            return FeatureIDEReader(filename).transform()
        elif filename.endswith(".afm"):
            return AFMReader(filename).transform()
        elif filename.endswith(".gfm.json"):
            return GlencoeReader(filename).transform()
        elif filename.endswith(".json"):
            return JSONReader(filename).transform()
    except Exception as e:
        raise FlamaException(f"Error reading feature model from {filename}: {e}")
    return None


def get_filename_from_url(url: str) -> str:
    """
    Extract the file name from a URL using pathlib.

    Example:
    - https://.../models/pizzas.uvl?token=XYZ -> pizzas
    """
    parsed = urlparse(url)
    path = pathlib.PurePosixPath(parsed.path)  # Use PurePosixPath to handle POSIX paths
    return path.name.split('.')[0]


from typing import Collection, Optional

from flamapy.core.exceptions import FlamaException
from flamapy.metamodels.fm_metamodel.models import FeatureModel
from flamapy.metamodels.fm_metamodel.transformations import (
    UVLReader, 
    FeatureIDEReader, 
    GlencoeReader,
    AFMReader,
    JSONReader
)


def int_to_scientific_notation(n: int, precision: int = 2) -> str:
    """Convert a large int into scientific notation.
    
    It is required for large numbers that Python cannot convert to float,
    solving the error `OverflowError: int too large to convert to float`.
    """
    str_n = str(n)
    decimal = str_n[1:precision+1]
    exponent = str(len(str_n) - 1)
    return str_n[0] + '.' + decimal + 'e' + exponent


def get_nof_configuration_as_str(nof_configurations: int, aproximation: bool, nof_cross_tree_constraints: int) -> str:
    return f"{'≤ ' if aproximation and nof_cross_tree_constraints > 0 else ''}{int_to_scientific_notation(nof_configurations) if nof_configurations > 1e6 else nof_configurations}"


def get_ratio(collection1: Collection, collection2: Collection, precision: int = 2) -> float:
    return 0.0 if not collection2 else round(len(collection1) / len(collection2), precision)


def get_percentage_str(value: int | float, precision: int = 4) -> str:
        if value == 0:
            return str(value)
        percentage = value * 100
        format_percentage = '{:.pe}'
        format_percentage = format_percentage.replace('p', str(precision))
        percentage_value = round(percentage, precision)
        return str(percentage_value) if percentage_value > 0 else format_percentage.format(percentage)


def read_fm_file(filename: str) -> Optional[FeatureModel]:
    try:
        if filename.endswith(".uvl"):
            return UVLReader(filename).transform()
        elif filename.endswith(".xml") or filename.endswith(".fide"):
            return FeatureIDEReader(filename).transform()
        elif filename.endswith("gfm.json"):
            return GlencoeReader(filename).transform()
        elif filename.endswith(".afm"):
            return AFMReader(filename).transform()
        elif filename.endswith(".json"):
            return JSONReader(filename).transform()
        else:
            raise FlamaException(f"Unsupported file format: {filename}")
    except Exception as e:
        raise FlamaException(f"Error reading feature model from {filename}: {e}")


from .fm_properties import FMProperty, FMPropertyMeasure, FMProperties
from .fm_metadata import FMMetadata
from .fm_metrics import FMMetrics
from .fm_analysis import FMAnalysis
from .characterization import FMCharacterization


__all__ = ['FMProperty', 'FMPropertyMeasure', 'FMProperties',
           'FMMetadata', 'FMMetrics', 'FMAnalysis',
           'FMCharacterization']


import math
import pathlib
import logging
from typing import Any

from fmfactlabel import FMProperties, FMPropertyMeasure
from .fm_utils import get_ratio, get_nof_configuration_as_str, get_percentage_str

from flamapy.metamodels.fm_metamodel.models import FeatureModel
from flamapy.metamodels.pysat_metamodel.transformations import FmToPysat
from flamapy.metamodels.bdd_metamodel.transformations import FmToBDD
from flamapy.metamodels.pysat_metamodel import operations as sat_operations
from flamapy.metamodels.bdd_metamodel import operations as bdd_operations
from flamapy.metamodels.fm_metamodel import operations as fm_operations


class FMAnalysis():

    def __init__(self, model: FeatureModel, light_fact_label: bool = False) -> None:
        self.fm = model
        self.light_fact_label = light_fact_label
        self.bdd_model = None
        self.sat_model = FmToPysat(model).transform()
        self.sat_model.original_model = self.fm
        if not self.light_fact_label:
            try:
                self.bdd_model = FmToBDD(model).transform()
            except Exception as e:
                logging.warning(f'Warning: the feature model is too large to build the BDD model. (Exception: {e})')

        # For performance purposes
        self._features = self.fm.get_features()
        if self.bdd_model is not None:
            self._configurations = bdd_operations.BDDConfigurationsNumber().execute(self.bdd_model).get_result()
            self._approximation = False
            self._fip = bdd_operations.BDDFeatureInclusionProbability().execute(self.bdd_model).get_result()
            self._pd = bdd_operations.BDDProductDistribution().execute(self.bdd_model).get_result()
            self._descriptive_statistics = descriptive_statistics(self._pd)
            self._core_features = [feat for feat, prob, in self._fip.items() if prob >= 1.0]
            self._dead_features = [feat for feat, prob, in self._fip.items() if prob <= 0.0]
            self._variant_features = [feat for feat, prob, in self._fip.items() if 0.0 < prob < 1.0]
        else:
            self._configurations = fm_operations.FMEstimatedConfigurationsNumber().execute(self.fm).get_result()
            self._approximation = True
            self._core_features = sat_operations.PySATCoreFeatures().execute(self.sat_model).get_result()
            self._dead_features = sat_operations.PySATDeadFeatures().execute(self.sat_model).get_result()
            self._variant_features = [f.name for f in self._features 
                                      if f.name not in self._core_features and
                                      f.name not in self._dead_features]
            self._fip = None
            self._descriptive_statistics = None

    def clean(self) -> None:
        if self.bdd_model is not None:
            logging.warning(f'BDD temp filepath: {self.bdd_model.bdd_file}')
            filepath = self.bdd_model.bdd_file
            filepath = filepath + '.dddmp' if not filepath.endswith('.dddmp') else filepath
            bdd_filepath = pathlib.Path(filepath)
            if bdd_filepath.exists():
                bdd_filepath.unlink()

    def get_analysis(self) -> list[FMPropertyMeasure]:
        result = []
        result.append(self.fm_valid())
        result.append(self.fm_core_features())
        result.append(self.fm_false_optional_features())
        result.append(self.fm_dead_features())
        result.append(self.fm_variant_features())
        if self.bdd_model is not None:
            result.append(self.fm_unique_features())
        if self._fip is not None:
            result.append(self.fm_pure_optional_features())
        result.append(self.fm_configurations_number())
        result.append(self.fm_total_variability())
        result.append(self.fm_partial_variability())
        if self.bdd_model is not None:
            result.append(self.fm_homogeneity())
        if self._descriptive_statistics is not None:
            result.append(self.fm_product_distribution())
            result.append(self.fm_mean_pd())
            result.append(self.fm_std_pd())
            result.append(self.fm_median_pd())
            result.append(self.fm_mad_pd())
            result.append(self.fm_mode_pd())
            result.append(self.fm_min_pd())
            result.append(self.fm_max_pd())
            result.append(self.fm_range_pd())
        return result

    def fm_valid(self) -> FMPropertyMeasure:
        if self.bdd_model is not None:
            _valid = self._configurations > 0
        else:
            _valid = sat_operations.PySATSatisfiable().execute(self.sat_model).get_result()
        _result = 'Yes' if _valid else 'No'
        return FMPropertyMeasure(FMProperties.VALID.value, _result)

    def fm_core_features(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.CORE_FEATURES.value,
                                 self._core_features, 
                                 len(self._core_features),
                                 get_ratio(self._core_features, self._features))

    def fm_dead_features(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.DEAD_FEATURES.value, 
                                 self._dead_features, 
                                 len(self._dead_features),
                                 get_ratio(self._dead_features, self._features))

    def fm_variant_features(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.VARIANT_FEATURES.value, 
                        self._variant_features, 
                        len(self._variant_features),
                        get_ratio(self._variant_features, self._features))
    
    def fm_unique_features(self) -> FMPropertyMeasure:
        _unique_features = bdd_operations.BDDUniqueFeatures().execute(self.bdd_model).get_result()
        return FMPropertyMeasure(FMProperties.UNIQUE_FEATURES.value, 
                                 _unique_features, 
                                 len(_unique_features),
                                 get_ratio(_unique_features, self._features))
    
    def fm_pure_optional_features(self) -> FMPropertyMeasure:
        _pure_optional_features = [feat for feat, prob, in self._fip.items() if prob == 0.5]
        return FMPropertyMeasure(FMProperties.PURE_OPTIONAL_FEATURES.value, 
                                 _pure_optional_features, 
                                 len(_pure_optional_features),
                                 get_ratio(_pure_optional_features, self._features))

    def fm_false_optional_features(self) -> FMPropertyMeasure:
        if self.bdd_model is not None:
            _false_optional_features = []
            for feat in self._core_features:
                feature = self.fm.get_feature_by_name(feat)
                if feature is not None and not feature.is_root() and not feature.is_mandatory():
                    _false_optional_features.append(feat)
        else:
            try:
                _false_optional_features = sat_operations.PySATFalseOptionalFeatures().execute(self.sat_model).get_result()
            except AssertionError as e:
                logging.warning(f'Warning: Feature model has feature cardinalities, false optional features cannot be computed.\n {e}')
                _false_optional_features = []
        return FMPropertyMeasure(FMProperties.FALSE_OPTIONAL_FEATURES.value, 
                                 _false_optional_features, 
                                 len(_false_optional_features),
                                 get_ratio(_false_optional_features, self._features))

    def fm_configurations_number(self) -> FMPropertyMeasure:
        _configurations = get_nof_configuration_as_str(self._configurations, self._approximation, len(self.fm.get_constraints()))
        return FMPropertyMeasure(FMProperties.CONFIGURATIONS.value, _configurations)
    
    def fm_total_variability(self) -> FMPropertyMeasure:
        _total_variability = self._configurations / (2 ** len(self._features) - 1)
        _total_variability = get_percentage_str(_total_variability, 2) + "%"
        return FMPropertyMeasure(FMProperties.TOTAL_VARIABILITY.value, _total_variability)
    
    def fm_partial_variability(self) -> FMPropertyMeasure:
        _partial_variability = 0 if not self._variant_features else self._configurations / (2 ** len(self._variant_features) - 1)
        _partial_variability = get_percentage_str(_partial_variability, 2) + "%"
        return FMPropertyMeasure(FMProperties.PARTIAL_VARIABILITY.value, _partial_variability)
    
    def fm_homogeneity(self) -> FMPropertyMeasure:
        _homogeneity = bdd_operations.BDDHomogeneity().execute(self.bdd_model).get_result()
        _homogeneity = get_percentage_str(_homogeneity, 2) + "%"
        return FMPropertyMeasure(FMProperties.HOMOGENEITY.value, _homogeneity)

    def fm_product_distribution(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.PRODUCT_DISTRIBUTION.value, None)

    def fm_mean_pd(self) -> FMPropertyMeasure:
        _mean_pd = round(self._descriptive_statistics['Mean'], 2)
        return FMPropertyMeasure(FMProperties.PD_MEAN.value, _mean_pd)
    
    def fm_std_pd(self) -> FMPropertyMeasure:
        _std_pd = round(self._descriptive_statistics['Standard deviation'], 2)
        return FMPropertyMeasure(FMProperties.PD_STD.value, _std_pd)
    
    def fm_median_pd(self) -> FMPropertyMeasure:
        _median_pd = round(self._descriptive_statistics['Median'], 2)
        return FMPropertyMeasure(FMProperties.PD_MEDIAN.value, _median_pd)
    
    def fm_mad_pd(self) -> FMPropertyMeasure:
        _mad_pd = round(self._descriptive_statistics['Median absolute deviation'], 2)
        return FMPropertyMeasure(FMProperties.PD_MAD.value, _mad_pd)
    
    def fm_mode_pd(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.PD_MODE.value, self._descriptive_statistics['Mode'])
    
    def fm_min_pd(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.PD_MIN.value, self._descriptive_statistics['Min'])
    
    def fm_max_pd(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.PD_MAX.value, self._descriptive_statistics['Max'])
    
    def fm_range_pd(self) -> FMPropertyMeasure:
        return FMPropertyMeasure(FMProperties.PD_RANGE.value, self._descriptive_statistics['Range'])


def descriptive_statistics(frequencies: list[int]) -> dict[str, Any]:
    total_count = sum(frequencies)
    
    # Mean calculation
    mean = sum(i * freq for i, freq in enumerate(frequencies)) / total_count
    
    # Standard deviation calculation
    variance = sum(freq * (i - mean) ** 2 for i, freq in enumerate(frequencies)) / total_count
    std_dev = math.sqrt(variance)
    
    # Median calculation
    cumulative_count = 0
    median_position = total_count / 2
    median = None
    for i, freq in enumerate(frequencies):
        cumulative_count += freq
        if cumulative_count >= median_position:
            median = i
            break
    
    # Median Absolute Deviation (MAD) calculation
    cumulative_count = 0
    mad_total = 0
    for i, freq in enumerate(frequencies):
        mad_total += freq * abs(i - median)
    mad = mad_total / total_count
    
    # Mode calculation
    mode_val = max(range(len(frequencies)), key=lambda i: frequencies[i])
    
    # Min and Max calculation
    min_val = next(i for i, freq in enumerate(frequencies) if freq > 0)
    max_val = next(i for i, freq in reversed(list(enumerate(frequencies))) if freq > 0)
    
    # Range calculation
    range_val = max_val - min_val


    return {
        'Mean': mean,
        'Standard deviation': std_dev,
        'Median': median,
        'Median absolute deviation': mad,
        'Mode': mode_val,
        'Min': min_val,
        'Max': max_val,
        'Range': range_val
    }


from typing import Optional

from fmfactlabel import FMProperties, FMPropertyMeasure

from flamapy.metamodels.fm_metamodel.models import FeatureModel
from flamapy.metamodels.fm_metamodel.operations import FMLanguageLevel


class FMMetadata():

    def __init__(self, 
                 model: FeatureModel, 
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 author: Optional[str] = None,
                 year: Optional[int] = None,
                 reference: Optional[str] = None,
                 tags: Optional[str] = None,
                 domains: Optional[list[str]] = None) -> None:
        self.fm = model
        self.name = name 
        self.description = description 
        self.author = author 
        self.year = year 
        self.reference = reference 
        self.tags = tags 
        self.domains = domains 

    def get_metadata(self) -> list[FMPropertyMeasure]:
        result = []
        result.append(self.fm_name(self.name))
        result.append(self.fm_description(self.description))
        result.append(self.fm_author(self.author))
        result.append(self.fm_year(self.year))
        result.append(self.fm_reference(self.reference))
        result.append(self.fm_tags(self.tags))
        result.append(self.fm_domains(self.domains))
        result.append(self.fm_language_level())
        return result

    def fm_name(self, value: Optional[str] = None) -> FMPropertyMeasure:
        value = value if value is not None else self.fm.root.name 
        return FMPropertyMeasure(FMProperties.NAME.value, value)
    
    def fm_description(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.DESCRIPTION.value)    
        return FMPropertyMeasure(FMProperties.DESCRIPTION.value, value)

    def fm_author(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.AUTHOR.value)    
        return FMPropertyMeasure(FMProperties.AUTHOR.value, value)
        
    def fm_year(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.YEAR.value)    
        return FMPropertyMeasure(FMProperties.YEAR.value, value)
    
    def fm_reference(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.REFERENCE.value)    
        return FMPropertyMeasure(FMProperties.REFERENCE.value, value)

    def fm_tags(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.TAGS.value)    
        return FMPropertyMeasure(FMProperties.TAGS.value, value)
    
    def fm_domains(self, value: Optional[str] = None) -> FMPropertyMeasure:
        if value is None:
            return FMPropertyMeasure(FMProperties.DOMAIN.value)    
        return FMPropertyMeasure(FMProperties.DOMAIN.value, value)
    
    def fm_language_level(self) -> FMPropertyMeasure:
        value = FMLanguageLevel().execute(self.fm).get_result()
        minor_levels = ', '.join(level.name.capitalize().replace('_', ' ') for level in value.minors)
        minor_levels = f' ({minor_levels})' if minor_levels else ''
        value = f'{value.major.name.capitalize()}{minor_levels}'
        return FMPropertyMeasure(FMProperties.LANGUAGE_LEVEL.value, value)



import statistics
from typing import Any
from collections import defaultdict

from fmfactlabel import FMProperties, FMPropertyMeasure
from .fm_utils import get_ratio
from flamapy.metamodels.fm_metamodel.models import FeatureModel, Feature, FeatureType


class FMMetrics():

    PRECISION: int = 2

    def __init__(self, model: FeatureModel):
        self.fm = model
        self._metrics: dict[str, Any] = traverse_metrics(self.fm)
                     
    def get_metrics(self) -> list[FMPropertyMeasure]:
        result = []
        result.append(self.fm_features())
        result.append(self.fm_abstract_features())
        result.append(self.fm_abstract_leaf_features())
        result.append(self.fm_abstract_compound_features())
        result.append(self.fm_concrete_features())
        result.append(self.fm_concrete_leaf_features())
        result.append(self.fm_concrete_compound_features())
        result.append(self.fm_compound_features())
        result.append(self.fm_leaf_features())
        result.append(self.fm_root_feature())
        result.append(self.fm_top_features())
        result.append(self.fm_solitary_features())
        result.append(self.fm_grouped_features())
        result.append(self.fm_typed_features())
        result.append(self.fm_numerical_features())
        result.append(self.fm_integer_features())
        result.append(self.fm_real_features())
        result.append(self.fm_string_features())
        result.append(self.fm_multi_features())
        result.append(self.fm_tree_relationships())
        result.append(self.fm_mandatory_features())
        result.append(self.fm_optional_features())
        result.append(self.fm_feature_groups())
        result.append(self.fm_alternative_groups())
        result.append(self.fm_or_groups())
        result.append(self.fm_mutex_groups())
        result.append(self.fm_cardinality_groups())
        result.append(self.fm_depth_tree())
        result.append(self.fm_mean_depth_tree())
        result.append(self.fm_avg_branching_factor())
        result.append(self.fm_min_children_per_feature())
        result.append(self.fm_max_children_per_feature())
        result.append(self.fm_avg_children_per_feature())
        result.append(self.fm_cross_tree_constraints())
        result.append(self.fm_logical_constraints())
        result.append(self.fm_single_feature_constraints())
        result.append(self.fm_simple_constraints())
        result.append(self.fm_requires_constraints())
        result.append(self.fm_excludes_constraints())
        result.append(self.fm_complex_constraints())
        result.append(self.fm_pseudocomplex_constraints())
        result.append(self.fm_strictcomplex_constraints())
        result.append(self.fm_arithmetic_constraints())
        result.append(self.fm_aggregation_constraints())
        result.append(self.fm_extra_constraint_representativeness())
        result.append(self.fm_min_features_per_constraint())
        result.append(self.fm_max_features_per_constraint())
        result.append(self.fm_avg_features_per_constraint())
        result.append(self.fm_avg_constraints_per_feature())
        result.append(self.fm_min_constraints_per_feature())
        result.append(self.fm_max_constraints_per_feature())
        result.append(self.fm_feature_attributes())
        result.append(self.fm_features_with_attributes())
        result.append(self.fm_min_attributes_per_feature())
        result.append(self.fm_max_attributes_per_feature())
        result.append(self.fm_avg_attributes_per_feature())
        result.append(self.fm_avg_attributes_per_feature_with_attributes())
        return result

    def fm_features(self) -> FMPropertyMeasure:
        _features = self._metrics[FMProperties.FEATURES.value]
        return FMPropertyMeasure(FMProperties.FEATURES.value, _features, len(_features))

    def fm_abstract_features(self) -> FMPropertyMeasure:
        _abstract_features = self._metrics[FMProperties.ABSTRACT_FEATURES.value]
        return FMPropertyMeasure(FMProperties.ABSTRACT_FEATURES.value, 
                                 _abstract_features, 
                                 len(_abstract_features),
                                 get_ratio(_abstract_features, self.fm_features().value))

    def fm_concrete_features(self) -> FMPropertyMeasure:
        _concrete_features = self._metrics[FMProperties.CONCRETE_FEATURES.value]
        return FMPropertyMeasure(FMProperties.CONCRETE_FEATURES.value, 
                                 _concrete_features, 
                                 len(_concrete_features),
                                 get_ratio(_concrete_features, self.fm_features().value))
    
    def fm_root_feature(self) -> FMPropertyMeasure:
        _root_feature = self._metrics[FMProperties.ROOT_FEATURE.value]
        return FMPropertyMeasure(FMProperties.ROOT_FEATURE.value, 
                                 _root_feature,
                                 len(_root_feature), 
                                 get_ratio(_root_feature, self.fm_features().value))

    def fm_top_features(self) -> FMPropertyMeasure:
        _top_features = self._metrics[FMProperties.TOP_FEATURES.value]
        return FMPropertyMeasure(FMProperties.TOP_FEATURES.value, 
                                 _top_features, 
                                 len(_top_features),
                                 get_ratio(_top_features, self.fm_features().value))

    def fm_leaf_features(self) -> FMPropertyMeasure:
        _leaf_features = self._metrics[FMProperties.LEAF_FEATURES.value]
        return FMPropertyMeasure(FMProperties.LEAF_FEATURES.value, 
                                 _leaf_features, 
                                 len(_leaf_features),
                                 get_ratio(_leaf_features, self.fm_features().value))

    def fm_compound_features(self) -> FMPropertyMeasure:
        _compound_features = self._metrics[FMProperties.COMPOUND_FEATURES.value]
        return FMPropertyMeasure(FMProperties.COMPOUND_FEATURES.value, 
                                 _compound_features, 
                                 len(_compound_features),
                                 get_ratio(_compound_features, self.fm_features().value))

    def fm_abstract_leaf_features(self) -> FMPropertyMeasure:
        _abstract_leaf_features = self._metrics[FMProperties.ABSTRACT_LEAF_FEATURES.value]
        return FMPropertyMeasure(FMProperties.ABSTRACT_LEAF_FEATURES.value, 
                                 _abstract_leaf_features, 
                                 len(_abstract_leaf_features),
                                 get_ratio(_abstract_leaf_features, self.fm_abstract_features().value))
    
    def fm_abstract_compound_features(self) -> FMPropertyMeasure:
        _abstract_compound_features = self._metrics[FMProperties.ABSTRACT_COMPOUND_FEATURES.value]
        return FMPropertyMeasure(FMProperties.ABSTRACT_COMPOUND_FEATURES.value, 
                                 _abstract_compound_features, 
                                 len(_abstract_compound_features),
                                 get_ratio(_abstract_compound_features, self.fm_abstract_features().value))

    def fm_concrete_leaf_features(self) -> FMPropertyMeasure:
        _concrete_leaf_features = self._metrics[FMProperties.CONCRETE_LEAF_FEATURES.value]
        return FMPropertyMeasure(FMProperties.CONCRETE_LEAF_FEATURES.value, 
                                 _concrete_leaf_features, 
                                 len(_concrete_leaf_features),
                                 get_ratio(_concrete_leaf_features, self.fm_concrete_features().value))
    
    def fm_concrete_compound_features(self) -> FMPropertyMeasure:
        _concrete_compound_features = self._metrics[FMProperties.CONCRETE_COMPOUND_FEATURES.value]
        return FMPropertyMeasure(FMProperties.CONCRETE_COMPOUND_FEATURES.value, 
                                 _concrete_compound_features, 
                                 len(_concrete_compound_features),
                                 get_ratio(_concrete_compound_features, self.fm_concrete_features().value))

    def fm_tree_relationships(self) -> FMPropertyMeasure:
        _tree_relationships = self._metrics[FMProperties.TREE_RELATIONSHIPS.value]
        return FMPropertyMeasure(FMProperties.TREE_RELATIONSHIPS.value, 
                                 _tree_relationships, 
                                 len(_tree_relationships))

    def fm_solitary_features(self) -> FMPropertyMeasure:
        _solitary_features = self._metrics[FMProperties.SOLITARY_FEATURES.value]
        return FMPropertyMeasure(FMProperties.SOLITARY_FEATURES.value, 
                                 _solitary_features, 
                                 len(_solitary_features),
                                 get_ratio(_solitary_features, self.fm_features().value))

    def fm_grouped_features(self) -> FMPropertyMeasure:
        _grouped_features = self._metrics[FMProperties.GROUPED_FEATURES.value]
        return FMPropertyMeasure(FMProperties.GROUPED_FEATURES.value, 
                                 _grouped_features, 
                                 len(_grouped_features),
                                 get_ratio(_grouped_features, self.fm_features().value))

    def fm_mandatory_features(self) -> FMPropertyMeasure:
        _mandatory_features = self._metrics[FMProperties.MANDATORY_FEATURES.value]
        return FMPropertyMeasure(FMProperties.MANDATORY_FEATURES.value, 
                                 _mandatory_features, 
                                 len(_mandatory_features),
                                 get_ratio(_mandatory_features, self.fm_solitary_features().value))

    def fm_optional_features(self) -> FMPropertyMeasure:
        _optional_features = self._metrics[FMProperties.OPTIONAL_FEATURES.value]
        return FMPropertyMeasure(FMProperties.OPTIONAL_FEATURES.value, 
                                 _optional_features, 
                                 len(_optional_features),
                                 get_ratio(_optional_features, self.fm_solitary_features().value))

    def fm_feature_groups(self) -> FMPropertyMeasure:
        _feature_groups = self._metrics[FMProperties.FEATURE_GROUPS.value]
        return FMPropertyMeasure(FMProperties.FEATURE_GROUPS.value, 
                                 _feature_groups, 
                                 len(_feature_groups),
                                 get_ratio(_feature_groups, self.fm_tree_relationships().value))
    
    def fm_alternative_groups(self) -> FMPropertyMeasure:
        _alternative_groups = self._metrics[FMProperties.ALTERNATIVE_GROUPS.value]
        return FMPropertyMeasure(FMProperties.ALTERNATIVE_GROUPS.value, 
                                 _alternative_groups, 
                                 len(_alternative_groups),
                                 get_ratio(_alternative_groups, self.fm_feature_groups().value))
    
    def fm_or_groups(self) -> FMPropertyMeasure:
        _or_groups = self._metrics[FMProperties.OR_GROUPS.value]
        return FMPropertyMeasure(FMProperties.OR_GROUPS.value, 
                                 _or_groups, 
                                 len(_or_groups),
                                 get_ratio(_or_groups, self.fm_feature_groups().value))

    def fm_mutex_groups(self) -> FMPropertyMeasure:
        _mutex_groups = self._metrics[FMProperties.MUTEX_GROUPS.value]
        return FMPropertyMeasure(FMProperties.MUTEX_GROUPS.value, 
                                 _mutex_groups, 
                                 len(_mutex_groups),
                                 get_ratio(_mutex_groups, self.fm_grouped_features().value))

    def fm_cardinality_groups(self) -> FMPropertyMeasure:
        _cardinality_groups = self._metrics[FMProperties.CARDINALITY_GROUPS.value]
        return FMPropertyMeasure(FMProperties.CARDINALITY_GROUPS.value, 
                                 _cardinality_groups, 
                                 len(_cardinality_groups),
                                 get_ratio(_cardinality_groups, self.fm_grouped_features().value))
    
    def fm_depth_tree(self) -> FMPropertyMeasure:
        _max_depth_tree = self._metrics[FMProperties.DEPTH_TREE.value]
        return FMPropertyMeasure(FMProperties.DEPTH_TREE.value, _max_depth_tree)

    def fm_mean_depth_tree(self) -> FMPropertyMeasure:
        _mean_depth_tree = self._metrics[FMProperties.MEAN_DEPTH_TREE.value]
        return FMPropertyMeasure(FMProperties.MEAN_DEPTH_TREE.value, round(_mean_depth_tree, FMMetrics.PRECISION))

    def fm_avg_branching_factor(self) -> FMPropertyMeasure:
        _avg_branching_factor = self._metrics[FMProperties.BRANCHING_FACTOR.value]
        return FMPropertyMeasure(FMProperties.BRANCHING_FACTOR.value, _avg_branching_factor)

    def fm_min_children_per_feature(self) -> FMPropertyMeasure:
        _min_children_per_feature = self._metrics[FMProperties.MIN_CHILDREN_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MIN_CHILDREN_PER_FEATURE.value, _min_children_per_feature)

    def fm_max_children_per_feature(self) -> FMPropertyMeasure:
        _max_children_per_feature = self._metrics[FMProperties.MAX_CHILDREN_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MAX_CHILDREN_PER_FEATURE.value, _max_children_per_feature)

    def fm_avg_children_per_feature(self) -> FMPropertyMeasure:
        _avg_children_per_feature = self._metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.AVG_CHILDREN_PER_FEATURE.value, round(_avg_children_per_feature, FMMetrics.PRECISION))

    def fm_cross_tree_constraints(self) -> FMPropertyMeasure:
        _cross_tree_constraints = self._metrics[FMProperties.CROSS_TREE_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.CROSS_TREE_CONSTRAINTS.value, 
                                 _cross_tree_constraints,
                                 len(_cross_tree_constraints))

    def fm_logical_constraints(self) -> FMPropertyMeasure:
        _logical_constraints = self._metrics[FMProperties.LOGICAL_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.LOGICAL_CONSTRAINTS.value, 
                                 _logical_constraints,
                                 len(_logical_constraints),
                                 get_ratio(_logical_constraints, self.fm_cross_tree_constraints().value))
    
    def fm_arithmetic_constraints(self) -> FMPropertyMeasure:
        _arithmetic_constraints = self._metrics[FMProperties.ARITHMETIC_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.ARITHMETIC_CONSTRAINTS.value, 
                                 _arithmetic_constraints,
                                 len(_arithmetic_constraints),
                                 get_ratio(_arithmetic_constraints, self.fm_cross_tree_constraints().value))
    
    def fm_aggregation_constraints(self) -> FMPropertyMeasure:
        _aggregation_constraints = self._metrics[FMProperties.AGGREGATION_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.AGGREGATION_CONSTRAINTS.value, 
                                 _aggregation_constraints,
                                 len(_aggregation_constraints),
                                 get_ratio(_aggregation_constraints, self.fm_cross_tree_constraints().value))
    
    def fm_single_feature_constraints(self) -> FMPropertyMeasure:
        _single_feature_constraints = self._metrics[FMProperties.SINGLE_FEATURE_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.SINGLE_FEATURE_CONSTRAINTS.value, 
                                 _single_feature_constraints,
                                 len(_single_feature_constraints),
                                 get_ratio(_single_feature_constraints, self.fm_logical_constraints().value))

    def fm_simple_constraints(self) -> FMPropertyMeasure:
        _simple_constraints = self._metrics[FMProperties.SIMPLE_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.SIMPLE_CONSTRAINTS.value, 
                                 _simple_constraints,
                                 len(_simple_constraints),
                                 get_ratio(_simple_constraints, self.fm_logical_constraints().value))

    def fm_requires_constraints(self) -> FMPropertyMeasure:
        _requires_constraints = self._metrics[FMProperties.REQUIRES_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.REQUIRES_CONSTRAINTS.value, 
                                 _requires_constraints,
                                 len(_requires_constraints),
                                 get_ratio(_requires_constraints, self.fm_simple_constraints().value))

    def fm_excludes_constraints(self) -> FMPropertyMeasure:
        _excludes_constraints = self._metrics[FMProperties.EXCLUDES_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.EXCLUDES_CONSTRAINTS.value, 
                                 _excludes_constraints,
                                 len(_excludes_constraints),
                                 get_ratio(_excludes_constraints, self.fm_simple_constraints().value))

    def fm_complex_constraints(self) -> FMPropertyMeasure:
        _complex_constraints = self._metrics[FMProperties.COMPLEX_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.COMPLEX_CONSTRAINTS.value, 
                                 _complex_constraints,
                                 len(_complex_constraints),
                                 get_ratio(_complex_constraints, self.fm_logical_constraints().value))

    def fm_pseudocomplex_constraints(self) -> FMPropertyMeasure:
        _pseudocomplex_constraints = self._metrics[FMProperties.PSEUDO_COMPLEX_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.PSEUDO_COMPLEX_CONSTRAINTS.value, 
                                 _pseudocomplex_constraints,
                                 len(_pseudocomplex_constraints),
                                 get_ratio(_pseudocomplex_constraints, self.fm_complex_constraints().value))

    def fm_strictcomplex_constraints(self) -> FMPropertyMeasure:
        _strictcomplex_constraints = self._metrics[FMProperties.STRICT_COMPLEX_CONSTRAINTS.value]
        return FMPropertyMeasure(FMProperties.STRICT_COMPLEX_CONSTRAINTS.value, 
                                 _strictcomplex_constraints,
                                 len(_strictcomplex_constraints),
                                 get_ratio(_strictcomplex_constraints, self.fm_complex_constraints().value))

    def fm_extra_constraint_representativeness(self) -> FMPropertyMeasure:
        _features_in_constraints = self._metrics[FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value]
        _ecr = get_ratio(_features_in_constraints, self.fm_features().value, FMMetrics.PRECISION)
        return FMPropertyMeasure(FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value,
                                 _features_in_constraints,
                                 len(_features_in_constraints),
                                 _ecr)

    def fm_min_constraints_per_feature(self) -> FMPropertyMeasure:
        _constraints_per_feature = self._metrics[FMProperties.MIN_CONSTRAINTS_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MIN_CONSTRAINTS_PER_FEATURE.value, _constraints_per_feature)

    def fm_max_constraints_per_feature(self) -> FMPropertyMeasure:
        _constraints_per_feature = self._metrics[FMProperties.MAX_CONSTRAINTS_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MAX_CONSTRAINTS_PER_FEATURE.value, _constraints_per_feature)

    def fm_avg_constraints_per_feature(self) -> FMPropertyMeasure:
        _constraints_per_feature = self._metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value, round(_constraints_per_feature, FMMetrics.PRECISION))
    
    def fm_min_features_per_constraint(self) -> FMPropertyMeasure:
        _features_per_constraint = self._metrics[FMProperties.MIN_FEATURES_PER_CONSTRAINT.value]
        return FMPropertyMeasure(FMProperties.MIN_FEATURES_PER_CONSTRAINT.value, _features_per_constraint)

    def fm_max_features_per_constraint(self) -> FMPropertyMeasure:
        _features_per_constraint = self._metrics[FMProperties.MAX_FEATURES_PER_CONSTRAINT.value]
        return FMPropertyMeasure(FMProperties.MAX_FEATURES_PER_CONSTRAINT.value, _features_per_constraint)

    def fm_avg_features_per_constraint(self) -> FMPropertyMeasure:
        _features_per_constraint = self._metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value]
        return FMPropertyMeasure(FMProperties.AVG_FEATURES_PER_CONSTRAINT.value, round(_features_per_constraint, FMMetrics.PRECISION))
    
    def fm_feature_attributes(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.FEATURE_ATTRIBUTES.value]
        return FMPropertyMeasure(FMProperties.FEATURE_ATTRIBUTES.value, _result, len(_result))
    
    def fm_features_with_attributes(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.FEATURES_WITH_ATTRIBUTES.value]
        return FMPropertyMeasure(FMProperties.FEATURES_WITH_ATTRIBUTES.value, _result, len(_result), 
                                 get_ratio(_result, self.fm_features().value))
    
    def fm_min_attributes_per_feature(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.MIN_ATTRIBUTES_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MIN_ATTRIBUTES_PER_FEATURE.value, _result)

    def fm_max_attributes_per_feature(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.MAX_ATTRIBUTES_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.MAX_ATTRIBUTES_PER_FEATURE.value, _result)

    def fm_avg_attributes_per_feature(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value]
        return FMPropertyMeasure(FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value, round(_result, FMMetrics.PRECISION))

    def fm_avg_attributes_per_feature_with_attributes(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value]
        return FMPropertyMeasure(FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value, round(_result, FMMetrics.PRECISION))

    def fm_multi_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.MULTI_FEATURES.value]
        return FMPropertyMeasure(FMProperties.MULTI_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))

    def fm_typed_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.TYPED_FEATURES.value]
        return FMPropertyMeasure(FMProperties.TYPED_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))

    def fm_numerical_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.NUMERICAL_FEATURES.value]
        return FMPropertyMeasure(FMProperties.NUMERICAL_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))

    def fm_integer_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.INTEGER_FEATURES.value]
        return FMPropertyMeasure(FMProperties.INTEGER_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))
    
    def fm_real_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.REAL_FEATURES.value]
        return FMPropertyMeasure(FMProperties.REAL_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))
    
    def fm_string_features(self) -> FMPropertyMeasure:
        _result = self._metrics[FMProperties.STRING_FEATURES.value]
        return FMPropertyMeasure(FMProperties.STRING_FEATURES.value, 
                                 _result,
                                 len(_result),
                                 get_ratio(_result, self.fm_features().value))
    

def traverse_metrics(fm: FeatureModel) -> dict[str, Any]:
    """Calculate all metrics from the feature model in only one traversing of the tree."""
    metrics: dict[str, Any] = {}
    if fm is None:
        return metrics
    ## Features metrics
    metrics[FMProperties.FEATURES.value] = list()
    metrics[FMProperties.ABSTRACT_FEATURES.value] = list()
    metrics[FMProperties.CONCRETE_FEATURES.value] = list()
    metrics[FMProperties.LEAF_FEATURES.value] = list()
    metrics[FMProperties.COMPOUND_FEATURES.value] = list()
    metrics[FMProperties.CONCRETE_COMPOUND_FEATURES.value] = list()
    metrics[FMProperties.CONCRETE_LEAF_FEATURES.value] = list()
    metrics[FMProperties.ABSTRACT_COMPOUND_FEATURES.value] = list()
    metrics[FMProperties.ABSTRACT_LEAF_FEATURES.value] = list()
    metrics[FMProperties.TREE_RELATIONSHIPS.value] = list()
    metrics[FMProperties.ROOT_FEATURE.value] = list()
    metrics[FMProperties.TOP_FEATURES.value] = list()
    metrics[FMProperties.SOLITARY_FEATURES.value] = list()
    metrics[FMProperties.GROUPED_FEATURES.value] = list()
    metrics[FMProperties.MANDATORY_FEATURES.value] = list()
    metrics[FMProperties.OPTIONAL_FEATURES.value] = list()
    metrics[FMProperties.FEATURE_GROUPS.value] = list()
    metrics[FMProperties.ALTERNATIVE_GROUPS.value] = list()
    metrics[FMProperties.OR_GROUPS.value] = list()
    metrics[FMProperties.MUTEX_GROUPS.value] = list()
    metrics[FMProperties.CARDINALITY_GROUPS.value] = list()
    metrics[FMProperties.MIN_CHILDREN_PER_FEATURE.value] = None
    metrics[FMProperties.MAX_CHILDREN_PER_FEATURE.value] = 0
    metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value] = list()
    metrics[FMProperties.DEPTH_TREE.value] = 0
    metrics[FMProperties.MEAN_DEPTH_TREE.value] = list()
    metrics[FMProperties.FEATURE_ATTRIBUTES.value] = set()
    metrics[FMProperties.FEATURES_WITH_ATTRIBUTES.value] = list()
    metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value] = list()
    metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value] = list()
    metrics['Branches'] = 0
    metrics['Children'] = 0
    metrics[FMProperties.INTEGER_FEATURES.value] = list()
    metrics[FMProperties.REAL_FEATURES.value] = list()
    metrics[FMProperties.STRING_FEATURES.value] = list()
    metrics[FMProperties.NUMERICAL_FEATURES.value] = list()
    metrics[FMProperties.TYPED_FEATURES.value] = list()
    metrics[FMProperties.MULTI_FEATURES.value] = list()

    traverse_feature_metrics(fm.root, metrics)
    metrics[FMProperties.BRANCHING_FACTOR.value] = 0 if metrics['Branches'] == 0 else round(metrics['Children'] / metrics['Branches'], FMMetrics.PRECISION)
    metrics[FMProperties.MEAN_DEPTH_TREE.value] = 0 if not metrics[FMProperties.MEAN_DEPTH_TREE.value] else statistics.mean(metrics[FMProperties.MEAN_DEPTH_TREE.value])
    metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value] else statistics.mean(metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value])
    metrics[FMProperties.MIN_ATTRIBUTES_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value] else min(metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value])
    metrics[FMProperties.MAX_ATTRIBUTES_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value] else max(metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value])
    metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value] else statistics.mean(metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value])
    metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value] = 0 if not metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value] else statistics.mean(metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value])
    metrics[FMProperties.FEATURE_ATTRIBUTES.value] = list(metrics[FMProperties.FEATURE_ATTRIBUTES.value])

    ## Constraints metrics
    metrics[FMProperties.CROSS_TREE_CONSTRAINTS.value] = list()
    metrics[FMProperties.LOGICAL_CONSTRAINTS.value] = list()
    metrics[FMProperties.ARITHMETIC_CONSTRAINTS.value] = list()
    metrics[FMProperties.AGGREGATION_CONSTRAINTS.value] = list()
    metrics[FMProperties.SINGLE_FEATURE_CONSTRAINTS.value] = list()
    metrics[FMProperties.SIMPLE_CONSTRAINTS.value] = list()
    metrics[FMProperties.REQUIRES_CONSTRAINTS.value] = list()
    metrics[FMProperties.EXCLUDES_CONSTRAINTS.value] = list()
    metrics[FMProperties.COMPLEX_CONSTRAINTS.value] = list()
    metrics[FMProperties.PSEUDO_COMPLEX_CONSTRAINTS.value] = list()
    metrics[FMProperties.STRICT_COMPLEX_CONSTRAINTS.value] = list()
    metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value] = defaultdict(int)
    metrics[FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value] = set()
    metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value] = list()

    traverse_constraints_metrics(fm, metrics)
    metrics[FMProperties.MIN_FEATURES_PER_CONSTRAINT.value] = 0 if not metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value] else min(metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value])
    metrics[FMProperties.MAX_FEATURES_PER_CONSTRAINT.value] = 0 if not metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value] else max(metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value])
    metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value] = 0 if not metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value] else statistics.mean(metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value])
    metrics[FMProperties.MIN_CONSTRAINTS_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value] else min(metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value].values())
    metrics[FMProperties.MAX_CONSTRAINTS_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value] else max(metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value].values())
    metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value] = 0 if not metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value] else statistics.mean(metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value].values())
    metrics[FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value] = list(metrics[FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value])
    return metrics


def traverse_feature_metrics(feature: Feature, metrics: dict[str, Any], depth: int = 0) -> None:
    if feature is not None:
        metrics[FMProperties.FEATURES.value].append(feature.name)
        if feature.parent is None:
            metrics[FMProperties.ROOT_FEATURE.value].append(feature.name)
            metrics[FMProperties.SOLITARY_FEATURES.value].append(feature.name)
        elif feature.parent.is_root():
            metrics[FMProperties.TOP_FEATURES.value].append(feature.name)
        if feature.parent is not None and not feature.parent.is_group():
            metrics[FMProperties.SOLITARY_FEATURES.value].append(feature.name)
        if feature.feature_cardinality.min != 1 or feature.feature_cardinality.max != 1:
            metrics[FMProperties.MULTI_FEATURES.value].append(feature.name)
        if feature.feature_type != FeatureType.BOOLEAN:
            metrics[FMProperties.TYPED_FEATURES.value].append(feature.name)
            if feature.feature_type == FeatureType.INTEGER:
                metrics[FMProperties.INTEGER_FEATURES.value].append(feature.name)
                metrics[FMProperties.NUMERICAL_FEATURES.value].append(feature.name)
            elif feature.feature_type == FeatureType.REAL:
                metrics[FMProperties.REAL_FEATURES.value].append(feature.name)
                metrics[FMProperties.NUMERICAL_FEATURES.value].append(feature.name)
            elif feature.feature_type == FeatureType.STRING:
                metrics[FMProperties.STRING_FEATURES.value].append(feature.name)

        # Attributes
        attributes = feature.get_attributes()
        metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE.value].append(len(attributes))
        if attributes:
            metrics[FMProperties.FEATURES_WITH_ATTRIBUTES.value].append(feature.name)
            metrics[FMProperties.AVG_ATTRIBUTES_PER_FEATURE_WITH_ATTRIBUTES.value].append(len(attributes))
            for attribute in attributes:
                metrics[FMProperties.FEATURE_ATTRIBUTES.value].add(attribute.name)

        relations = feature.get_relations()
        n_children = 0
        if relations:  # it is a compound feature (non leaf)
            metrics[FMProperties.COMPOUND_FEATURES.value].append(feature.name)
            if feature.is_abstract:
                metrics[FMProperties.ABSTRACT_FEATURES.value].append(feature.name)
                metrics[FMProperties.ABSTRACT_COMPOUND_FEATURES.value].append(feature.name)
            else:
                metrics[FMProperties.CONCRETE_FEATURES.value].append(feature.name)
                metrics[FMProperties.CONCRETE_COMPOUND_FEATURES.value].append(feature.name)
            metrics['Branches'] += 1

            for relation in feature.get_relations():
                metrics[FMProperties.TREE_RELATIONSHIPS.value].append(str(relation))
                if relation.is_mandatory():
                    n_children += 1
                    metrics[FMProperties.MANDATORY_FEATURES.value].append(feature.name)
                    traverse_feature_metrics(relation.children[0], metrics, depth + 1)
                elif relation.is_optional():
                    n_children += 1
                    metrics[FMProperties.OPTIONAL_FEATURES.value].append(feature.name)
                    traverse_feature_metrics(relation.children[0], metrics, depth + 1)
                elif relation.is_group():
                    n_children += len(relation.children)
                    metrics[FMProperties.FEATURE_GROUPS.value].append(str(relation))
                    if relation.is_or():
                        metrics[FMProperties.OR_GROUPS.value].append(str(relation))
                    elif relation.is_alternative():
                        metrics[FMProperties.ALTERNATIVE_GROUPS.value].append(str(relation))
                    elif relation.is_mutex():
                        metrics[FMProperties.MUTEX_GROUPS.value].append(str(relation))
                    else:
                        metrics[FMProperties.CARDINALITY_GROUPS.value].append(str(relation))
                    for child in relation.children:
                        metrics[FMProperties.GROUPED_FEATURES.value].append(child.name)
                        traverse_feature_metrics(child, metrics, depth + 1)
            metrics['Children'] += n_children
            metrics[FMProperties.MIN_CHILDREN_PER_FEATURE.value] = n_children if metrics[FMProperties.MIN_CHILDREN_PER_FEATURE.value] is None else min(metrics[FMProperties.MIN_CHILDREN_PER_FEATURE.value], n_children)
            metrics[FMProperties.MAX_CHILDREN_PER_FEATURE.value] = max(metrics[FMProperties.MAX_CHILDREN_PER_FEATURE.value], n_children)
        else:  # it is a leaf feature
            metrics[FMProperties.LEAF_FEATURES.value].append(feature.name)
            if feature.is_abstract:
                metrics[FMProperties.ABSTRACT_FEATURES.value].append(feature.name)
                metrics[FMProperties.ABSTRACT_LEAF_FEATURES.value].append(feature.name)
            else:
                metrics[FMProperties.CONCRETE_FEATURES.value].append(feature.name)
                metrics[FMProperties.CONCRETE_LEAF_FEATURES.value].append(feature.name)
            metrics[FMProperties.DEPTH_TREE.value] = max(metrics[FMProperties.DEPTH_TREE.value], depth)
            metrics[FMProperties.MEAN_DEPTH_TREE.value].append(depth)
        metrics[FMProperties.AVG_CHILDREN_PER_FEATURE.value].append(n_children)


def traverse_constraints_metrics(fm: FeatureModel, metrics: dict[str, Any]) -> None:
    for ctc in fm.get_constraints():
        metrics[FMProperties.CROSS_TREE_CONSTRAINTS.value].append(ctc.ast.pretty_str())
        if ctc.is_logical_constraint():
            metrics[FMProperties.LOGICAL_CONSTRAINTS.value].append(ctc.ast.pretty_str())
            if ctc.is_single_feature_constraint():
                metrics[FMProperties.SINGLE_FEATURE_CONSTRAINTS.value].append(ctc.ast.pretty_str())
            elif ctc.is_requires_constraint():
                metrics[FMProperties.SIMPLE_CONSTRAINTS.value].append(ctc.ast.pretty_str())
                metrics[FMProperties.REQUIRES_CONSTRAINTS.value].append(ctc.ast.pretty_str())
            elif ctc.is_excludes_constraint():
                metrics[FMProperties.SIMPLE_CONSTRAINTS.value].append(ctc.ast.pretty_str())
                metrics[FMProperties.EXCLUDES_CONSTRAINTS.value].append(ctc.ast.pretty_str())
            else:
                metrics[FMProperties.COMPLEX_CONSTRAINTS.value].append(ctc.ast.pretty_str())
                if ctc.is_pseudocomplex_constraint():
                    metrics[FMProperties.PSEUDO_COMPLEX_CONSTRAINTS.value].append(ctc.ast.pretty_str())
                else:
                    metrics[FMProperties.STRICT_COMPLEX_CONSTRAINTS.value].append(ctc.ast.pretty_str())
        elif ctc.is_aggregation_constraint():
            metrics[FMProperties.AGGREGATION_CONSTRAINTS.value].append(ctc.ast.pretty_str())
        elif ctc.is_arithmetic_constraint():
            metrics[FMProperties.ARITHMETIC_CONSTRAINTS.value].append(ctc.ast.pretty_str())
        
        features = ctc.get_features()
        metrics[FMProperties.EXTRA_CONSTRAINT_REPRESENTATIVENESS.value].update(features)
        metrics[FMProperties.AVG_FEATURES_PER_CONSTRAINT.value].append(len(features))
        for feature in features:
            metrics[FMProperties.AVG_CONSTRAINTS_PER_FEATURE.value][feature] += 1
